# Copyright 2016-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.baseclasses import MainHierarchicalParser
import numpy as np
from .commonparser import CPMDCommonParser
import re
import logging
LOGGER = logging.getLogger("nomad")


class CPMDGeoOptParser(MainHierarchicalParser):
    """The main parser class that is called for all run types. Parses the CPMD
    output file.
    """
    def __init__(self, parser_context):
        """
        """
        super(CPMDGeoOptParser, self).__init__(parser_context)
        self.setup_common_matcher(CPMDCommonParser(parser_context))
        self.n_frames = 0
        self.sampling_method_gid = None
        self.frame_refs = []
        self.energies = []

        #=======================================================================
        # Cache levels
        # self.caching_levels.update({
            # 'section_run': CachingLevel.ForwardAndCache,
        # })

        #=======================================================================
        # Main structure
        self.root_matcher = SM("",
            forwardMatch=True,
            sections=['section_run', "section_frame_sequence", "section_sampling_method",  "section_method"],
            subMatchers=[
                self.cm.header(),
                self.cm.method(),
                self.cm.atoms(),
                self.cm.cell(),
                self.cm.initialization(),
                SM( " INITIALIZE EMPIRICAL HESSIAN",
                    sections=["x_cpmd_section_geo_opt_initialization"],
                    subMatchers=[
                        SM("                           <<<<< ASSUMED BONDS >>>>>"),
                        SM("\s+{0}\s+<-->\s+{0}".format(self.regexs.int),
                            repeats=True
                        ),
                        SM(" TOTAL NUMBER OF MOLECULAR STRUCTURES:\s+(?P<x_cpmd_total_number_of_molecular_structures>{})".format(self.regexs.int)),
                        SM(re.escape("   ATOM          COORDINATES            GRADIENTS (-FORCES)"),
                            adHoc=self.cm.ad_hoc_positions_forces("x_cpmd_initialized_positions", "x_cpmd_initialized_forces", "bohr", "hartree/bohr"),
                        ),
                        SM(" CPU TIME FOR INITIALIZATION\s+(?P<x_cpmd_geo_opt_initialization_time>{}) SECONDS".format(self.regexs.float)),
                    ]
                ),
                SM( re.escape(" =                  GEOMETRY OPTIMIZATION                       ="),
                    endReStr=re.escape(" =              END OF GEOMETRY OPTIMIZATION                    ="),
                    subMatchers=[
                        SM(" NFI      GEMAX       CNORM           ETOT        DETOT      TCPU"),
                        SM(" EWALD\| SUM IN REAL SPACE OVER"),
                        SM("\s+{0}\s+{1}\s+{1}\s+{1}\s+{1}\s+{1}".format(self.regexs.int, self.regexs.float),
                            forwardMatch=True,
                            endReStr=re.escape(" *** CNSTR="),
                            repeats=True,
                            sections=["section_single_configuration_calculation", "section_system", "x_cpmd_section_geo_opt_step"],
                            subMatchers=[
                                SM( "\s+(?P<x_cpmd_geo_opt_scf_nfi>{0})\s+(?P<x_cpmd_geo_opt_scf_gemax>{1})\s+(?P<x_cpmd_geo_opt_scf_cnorm>{1})\s+(?P<x_cpmd_geo_opt_scf_etot__hartree>{1})\s+(?P<x_cpmd_geo_opt_scf_detot__hartree>{1})\s+(?P<x_cpmd_geo_opt_scf_tcpu__s>{1})".format(self.regexs.int, self.regexs.float),
                                    sections=["x_cpmd_section_geo_opt_scf_iteration"],
                                    repeats=True,
                                ),
                                SM(" RESTART INFORMATION WRITTEN ON FILE\s+{}".format(self.regexs.eol)),
                                SM( re.escape("   ATOM          COORDINATES            GRADIENTS (-FORCES)"),
                                    adHoc=self.cm.ad_hoc_positions_forces("x_cpmd_geo_opt_step_positions", "x_cpmd_geo_opt_step_forces", "bohr", "hartree/bohr"),
                                ),
                                SM(" *** TOTAL STEP NR\.\s+(?P<x_cpmd_geo_opt_step_total_number_of_scf_steps>{0})\s+GEOMETRY STEP NR\.\s+(?P<x_cpmd_geo_opt_step_number>{0})  ***".replace("*", "\*").format(self.regexs.int)),
                                SM(" *** GNMAX=\s+(?P<x_cpmd_geo_opt_step_gnmax>{0})(?:\s+\[{0}\])?\s+ETOT=\s+(?P<x_cpmd_geo_opt_step_etot__hartree>{0})  ***".replace("*", "\*").format(self.regexs.float)),
                                SM(" *** GNORM=\s+(?P<x_cpmd_geo_opt_step_gnorm>{0})\s+DETOT=\s+(?P<x_cpmd_geo_opt_step_detot>{0})  ***".replace("*", "\*").format(self.regexs.float)),
                                SM(" *** CNSTR=\s+(?P<x_cpmd_geo_opt_step_cnstr>{0})\s+TCPU=\s+(?P<x_cpmd_geo_opt_step_tcpu>{0})  ***".replace("*", "\*").format(self.regexs.float)),
                            ]
                        ),
                    ]
                ),
                self.cm.footer(),
            ]
        )

    #=======================================================================
    # onClose triggers
    def onClose_x_cpmd_section_geo_opt_scf_iteration(self, backend, gIndex, section):
        # SCF step energy and energy change
        scf_id = backend.openSection("section_scf_iteration")
        energy = section.get_latest_value("x_cpmd_geo_opt_scf_etot")
        backend.addValue("energy_total_scf_iteration", energy)
        denergy = section.get_latest_value("x_cpmd_geo_opt_scf_detot")
        backend.addValue("energy_change_scf_iteration", denergy)
        backend.closeSection("section_scf_iteration", scf_id)

    def onClose_x_cpmd_section_geo_opt_step(self, backend, gIndex, section):
        total_energy = section.get_latest_value("x_cpmd_geo_opt_step_etot")
        if total_energy:
            backend.addValue("energy_total", total_energy)
            self.energies.append(total_energy)
        forces = section.get_latest_value("x_cpmd_geo_opt_step_forces")
        backend.addArrayValues("atom_forces", forces)
        positions = section.get_latest_value("x_cpmd_geo_opt_step_positions")
        backend.addArrayValues("atom_positions", positions)
        self.n_frames += 1

    def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
        self.frame_refs.append(gIndex)

    def onClose_section_frame_sequence(self, backend, gIndex, section):
        backend.addValue("number_of_frames_in_sequence", self.n_frames)
        if self.sampling_method_gid is not None:
            backend.addValue("frame_sequence_to_sampling_ref", self.sampling_method_gid)
        if self.frame_refs:
            backend.addArrayValues("frame_sequence_local_frames_ref", np.array(self.frame_refs))
        if self.energies:
            backend.addArrayValues("frame_sequence_potential_energy", np.array(self.energies))

    def onClose_section_sampling_method(self, backend, gIndex, section):
        # For single point calculations there is only one method and system.
        self.sampling_method_gid = gIndex
        backend.addValue("sampling_method", "geometry_optimization")

    def onClose_section_system(self, backend, gIndex, section):
        self.cache_service.addArrayValues("atom_labels")
        self.cache_service.addArrayValues("simulation_cell", unit="bohr")
        self.cache_service.addValue("number_of_atoms")

    #=======================================================================
    # adHoc



import json
import os
import re
from typing import Union
import sys

import requests


REPOSITORY_URL = (
    "https://api.github.com/repos/Materials-Consortia/optimade-python-tools"
)


def get_json(request: requests.Response) -> Union[dict, list]:
    """Utility function to 'safely' get JSON from response"""
    try:
        response: Union[dict, list] = request.json()
    except json.JSONDecodeError:
        sys.exit(
            f"Could not decode '{request.url}'. Status code: {request.status_code}"
        )
    else:
        return response


# Check current GitHub ref is a valid version tag
GITHUB_REF = os.environ.get("GITHUB_REF", None)

if GITHUB_REF is None:
    sys.exit("Could not retrieve current GitHub ref from environment variable")

if re.match(r"refs/tags/v[0-9]+(\.[0-9]+){2}$", GITHUB_REF) is None:
    sys.exit(
        f"You are NOT publishing a tag with the correct name format (vMAJOR.MINOR.PATCH), but instead '{GITHUB_REF}', will shut the publishing workflow down"
    )


# Get latest release (latest tag)
all_tags: list = get_json(requests.get(f"{REPOSITORY_URL}/tags"))
latest_release: str = sorted([tag.get("name", "v0.0.0") for tag in all_tags])[-1]


# Get commit date and time for latest release
latest_release_commit: dict = get_json(
    requests.get(f"{REPOSITORY_URL}/commits/{latest_release}")
)
latest_release_datetime: Union[str, None] = latest_release_commit.get("commit", {}).get(
    "committer", {}
).get("date", None)

if latest_release_datetime is None:
    sys.exit(
        f"Could not retrieve the datetime for the latest release commit:\n\n{json.dumps(latest_release_commit, indent=2)}"
    )


# Get list of all commits since latest release in `master`
master_commits: list = get_json(
    requests.get(f"{REPOSITORY_URL}/commits?since={latest_release_datetime}")
)
master_commit_shas = [commit.get("sha", "") for commit in master_commits]


# Check current GitHub SHA is already in the `master` branch
GITHUB_SHA = os.environ.get("GITHUB_SHA", None)

if GITHUB_SHA is None:
    sys.exit("Could not retrieve current GitHub SHA from environment variable")

if GITHUB_SHA not in master_commit_shas:
    sys.exit(
        "You are NOT publishing a tag pointing to a commit in the `master` branch, will shut the publishing workflow down"
    )

print(
    "All seems to be in order:\n"
    f"The published tag has the correct type of name (vMAJOR.MINOR.PATCH: {GITHUB_REF[len('refs/tags/'):]}) and refers to an existing commit in the `master` branch: {GITHUB_SHA}."
)
sys.exit(0)



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='common_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='common.nomadmetainfo.json'))


class settings_atom_in_molecule(MCategory):
    '''
    Parameters of an atom within a molecule.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='settings_atom_in_molecule'))


class settings_constraint(MCategory):
    '''
    Some parameters that describe a constraint
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='settings_constraint'))


class settings_interaction(MCategory):
    '''
    Some parameters that describe a bonded interaction.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='settings_interaction'))


class soap_parameter(MCategory):
    '''
    A soap parameter
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='soap_parameter'))


class response_context(MSection):
    '''
    The top level context containing the reponse to an api query, when using jsonAPI they
    are tipically in the meta part
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='response_context'))

    shortened_meta_info = Quantity(
        type=str,
        shape=[],
        description='''
        A meta info whose corresponding data has been shortened
        ''',
        a_legacy=LegacyDefinition(name='shortened_meta_info'))

    section_response_message = SubSection(
        sub_section=SectionProxy('section_response_message'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_response_message'))


class section_atom_type(MSection):
    '''
    Section describing a type of atom in the system.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_atom_type'))

    atom_type_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='coulomb',
        description='''
        Charge of the atom type.
        ''',
        a_legacy=LegacyDefinition(name='atom_type_charge'))

    atom_type_mass = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kilogram',
        description='''
        Mass of the atom type.
        ''',
        a_legacy=LegacyDefinition(name='atom_type_mass'))

    atom_type_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name (label) of the atom type.
        ''',
        a_legacy=LegacyDefinition(name='atom_type_name'))


class section_constraint(MSection):
    '''
    Section describing a constraint between arbitrary atoms.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_constraint'))

    constraint_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_constraints', 'number_of_atoms_per_constraint'],
        description='''
        List of the indexes involved in this constraint. The fist atom has index 1, the
        last number_of_topology_atoms.
        ''',
        a_legacy=LegacyDefinition(name='constraint_atoms'))

    constraint_kind = Quantity(
        type=str,
        shape=[],
        description='''
        Short and unique name for this constraint type. Valid names are described in the
        [constraint\\_kind wiki page](https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-
        info/wikis/metainfo/constraint-kind).
        ''',
        a_legacy=LegacyDefinition(name='constraint_kind'))

    constraint_parameters = Quantity(
        type=typing.Any,
        shape=[],
        description='''
        Explicit constraint parameters for this kind of constraint (depending on the
        constraint type, some might be given implicitly through other means).
        ''',
        a_legacy=LegacyDefinition(name='constraint_parameters'))

    number_of_atoms_per_constraint = Quantity(
        type=int,
        shape=[],
        description='''
        Number of atoms involved in this constraint.
        ''',
        a_legacy=LegacyDefinition(name='number_of_atoms_per_constraint'))

    number_of_constraints = Quantity(
        type=int,
        shape=[],
        description='''
        Number of constraints of this type.
        ''',
        a_legacy=LegacyDefinition(name='number_of_constraints'))


class section_dft_plus_u_orbital(MSection):
    '''
    Section for DFT+U-settings of a single orbital
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_dft_plus_u_orbital'))

    dft_plus_u_orbital_atom = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        DFT+U-orbital setting: atom index (references index of atom_labels/atom_positions)
        ''',
        a_legacy=LegacyDefinition(name='dft_plus_u_orbital_atom'))

    dft_plus_u_orbital_J = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        DFT+U-orbital setting: value J (exchange interaction)
        ''',
        categories=[public.energy_value],
        a_legacy=LegacyDefinition(name='dft_plus_u_orbital_J'))

    dft_plus_u_orbital_label = Quantity(
        type=str,
        shape=[],
        description='''
        DFT+U-orbital setting: orbital label (normally (n,l)), notation: '3d', '4f', ...
        ''',
        a_legacy=LegacyDefinition(name='dft_plus_u_orbital_label'))

    dft_plus_u_orbital_U_effective = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        DFT+U-orbital setting: value U_{effective} (U-J), if implementation uses it
        ''',
        categories=[public.energy_value],
        a_legacy=LegacyDefinition(name='dft_plus_u_orbital_U_effective'))

    dft_plus_u_orbital_U = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        DFT+U-orbital setting: value U (on-site Coulomb interaction)
        ''',
        categories=[public.energy_value],
        a_legacy=LegacyDefinition(name='dft_plus_u_orbital_U'))


class section_excited_states(MSection):
    '''
    Excited states properties.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_excited_states'))

    excitation_energies = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_excited_states'],
        description='''
        Excitation energies.
        ''',
        categories=[public.energy_value],
        a_legacy=LegacyDefinition(name='excitation_energies'))

    number_of_excited_states = Quantity(
        type=int,
        shape=[],
        description='''
        Number of excited states.
        ''',
        a_legacy=LegacyDefinition(name='number_of_excited_states'))

    oscillator_strengths = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_excited_states'],
        description='''
        Excited states oscillator strengths.
        ''',
        a_legacy=LegacyDefinition(name='oscillator_strengths'))

    transition_dipole_moments = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_excited_states', 3],
        description='''
        Transition dipole moments.
        ''',
        a_legacy=LegacyDefinition(name='transition_dipole_moments'))


class section_interaction(MSection):
    '''
    Section containing the description of a bonded interaction between arbitrary atoms.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_interaction'))

    interaction_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_interactions', 'number_of_atoms_per_interaction'],
        description='''
        List of the indexes involved in this interaction. The fist atom has index 1, the
        last atom index number_of_topology_atoms.
        ''',
        a_legacy=LegacyDefinition(name='interaction_atoms'))

    interaction_kind = Quantity(
        type=str,
        shape=[],
        description='''
        Short and unique name for this interaction type. Valid names are described in the
        [interaction\\_kind wiki page](https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-
        info/wikis/metainfo/interaction-kind).
        ''',
        a_legacy=LegacyDefinition(name='interaction_kind'))

    interaction_parameters = Quantity(
        type=typing.Any,
        shape=[],
        description='''
        Explicit interaction parameters for this kind of interaction (depending on the
        interaction_kind some might be given implicitly through other means).
        ''',
        a_legacy=LegacyDefinition(name='interaction_parameters'))

    number_of_atoms_per_interaction = Quantity(
        type=int,
        shape=[],
        description='''
        Number of atoms involved in this interaction.
        ''',
        a_legacy=LegacyDefinition(name='number_of_atoms_per_interaction'))

    number_of_interactions = Quantity(
        type=int,
        shape=[],
        description='''
        Number of interactions of this type.
        ''',
        a_legacy=LegacyDefinition(name='number_of_interactions'))


class section_method_basis_set(MSection):
    '''
    This section contains the definition of the basis sets that are defined independently
    of the atomic configuration.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_method_basis_set'))

    mapping_section_method_basis_set_atom_centered = Quantity(
        type=np.dtype(np.int64),
        shape=['number_of_basis_sets_atom_centered', 2],
        description='''
        Reference to an atom-centered basis set defined in section_basis_set_atom_centered
        and to the atom kind as defined in section_method_atom_kind.
        ''',
        a_legacy=LegacyDefinition(name='mapping_section_method_basis_set_atom_centered'))

    mapping_section_method_basis_set_cell_associated = Quantity(
        type=public.section_basis_set_cell_dependent,
        shape=[],
        description='''
        Reference to a cell-associated basis set.
        ''',
        a_legacy=LegacyDefinition(name='mapping_section_method_basis_set_cell_associated'))

    method_basis_set_kind = Quantity(
        type=str,
        shape=[],
        description='''
        String describing the use of the basis set, i.e, if it used for expanding a
        wavefunction or an electron density. Allowed values are listed in the
        [basis\\_set\\_kind wiki page](https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-
        info/wikis/metainfo/basis-set-kind).
        ''',
        a_legacy=LegacyDefinition(name='method_basis_set_kind'))

    number_of_basis_sets_atom_centered = Quantity(
        type=int,
        shape=[],
        description='''
        String describing the use of the basis set, i.e, if it used for expanding a
        wavefunction or an electron density. Allowed values are listed in the
        [basis\\_set\\_kind wiki page](https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-
        info/wikis/metainfo/basis-set-kind).
        ''',
        a_legacy=LegacyDefinition(name='number_of_basis_sets_atom_centered'))


class section_molecule_constraint(MSection):
    '''
    Section describing a constraint between atoms within a molecule.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_molecule_constraint'))

    molecule_constraint_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_molecule_constraints', 'number_of_atoms_per_molecule_constraint'],
        description='''
        List of the indexes involved in this constraint. The fist atom has index 1, the
        last index is number_of_atoms_in_molecule.
        ''',
        a_legacy=LegacyDefinition(name='molecule_constraint_atoms'))

    molecule_constraint_kind = Quantity(
        type=str,
        shape=[],
        description='''
        Short and unique name for this constraint type. Valid names are described in the
        [constraint\\_kind wiki page](https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-
        info/wikis/metainfo/constraint-kind).
        ''',
        a_legacy=LegacyDefinition(name='molecule_constraint_kind'))

    molecule_constraint_parameters = Quantity(
        type=typing.Any,
        shape=[],
        description='''
        Explicit constraint parameters for this kind of constraint (depending on the
        constraint type some might be given implicitly through other means).
        ''',
        a_legacy=LegacyDefinition(name='molecule_constraint_parameters'))

    number_of_atoms_per_molecule_constraint = Quantity(
        type=int,
        shape=[],
        description='''
        Number of atoms, in this molecule, involved in this constraint.
        ''',
        a_legacy=LegacyDefinition(name='number_of_atoms_per_molecule_constraint'))

    number_of_molecule_constraints = Quantity(
        type=int,
        shape=[],
        description='''
        Number of constraints of this type.
        ''',
        a_legacy=LegacyDefinition(name='number_of_molecule_constraints'))


class section_molecule_interaction(MSection):
    '''
    Section describing a bonded interaction between atoms within a molecule.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_molecule_interaction'))

    molecule_interaction_atoms = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_molecule_interactions', 'number_of_atoms_per_molecule_interaction'],
        description='''
        List of the indexes involved in this bonded interaction within a molecule. The
        first atom has index 1, the last index is number_of_atoms_in_.
        ''',
        a_legacy=LegacyDefinition(name='molecule_interaction_atoms'))

    molecule_interaction_kind = Quantity(
        type=str,
        shape=[],
        description='''
        Short and unique name for this interaction type, used for bonded interactions for
        atoms in a molecule. Valid names are described in the [interaction\\_kind wiki
        page](https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-
        info/wikis/metainfo/interaction-kind).
        ''',
        a_legacy=LegacyDefinition(name='molecule_interaction_kind'))

    molecule_interaction_parameters = Quantity(
        type=typing.Any,
        shape=[],
        description='''
        Explicit interaction parameters for this kind of interaction (depending on the
        interaction type some might be given implicitly through other means), used for
        bonded interactions for atoms in a molecule.
        ''',
        a_legacy=LegacyDefinition(name='molecule_interaction_parameters'))

    number_of_atoms_per_molecule_interaction = Quantity(
        type=int,
        shape=[],
        description='''
        Number of atoms, in this molecule, involved in this interaction.
        ''',
        a_legacy=LegacyDefinition(name='number_of_atoms_per_molecule_interaction'))

    number_of_molecule_interactions = Quantity(
        type=int,
        shape=[],
        description='''
        Number of bonded interactions of this type.
        ''',
        a_legacy=LegacyDefinition(name='number_of_molecule_interactions'))


class section_molecule_type(MSection):
    '''
    Section describing a type of molecule in the system.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_molecule_type'))

    atom_in_molecule_charge = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms_in_molecule'],
        unit='coulomb',
        description='''
        Charge of each atom in the molecule.
        ''',
        categories=[settings_atom_in_molecule],
        a_legacy=LegacyDefinition(name='atom_in_molecule_charge'))

    atom_in_molecule_name = Quantity(
        type=str,
        shape=['number_of_atoms_in_molecule'],
        description='''
        Name (label) of each atom in the molecule.
        ''',
        categories=[settings_atom_in_molecule],
        a_legacy=LegacyDefinition(name='atom_in_molecule_name'))

    atom_in_molecule_to_atom_type_ref = Quantity(
        type=Reference(SectionProxy('section_atom_type')),
        shape=['number_of_atoms_in_molecule'],
        description='''
        Reference to the atom type of each atom in the molecule.
        ''',
        categories=[settings_atom_in_molecule],
        a_legacy=LegacyDefinition(name='atom_in_molecule_to_atom_type_ref'))

    molecule_type_name = Quantity(
        type=str,
        shape=[],
        description='''
        Name of the molecule.
        ''',
        a_legacy=LegacyDefinition(name='molecule_type_name'))

    number_of_atoms_in_molecule = Quantity(
        type=int,
        shape=[],
        description='''
        Number of atoms in this molecule.
        ''',
        a_legacy=LegacyDefinition(name='number_of_atoms_in_molecule'))

    section_molecule_constraint = SubSection(
        sub_section=SectionProxy('section_molecule_constraint'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_molecule_constraint'))

    section_molecule_interaction = SubSection(
        sub_section=SectionProxy('section_molecule_interaction'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_molecule_interaction'))


class section_response_message(MSection):
    '''
    Messages outputted by the program formatting the data in the current response
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_response_message'))

    response_message_count = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        How many times this message was repeated
        ''',
        a_legacy=LegacyDefinition(name='response_message_count'))

    response_message_level = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        level of the message: 0 fatal, 1 error, 2 warning, 3 debug
        ''',
        a_legacy=LegacyDefinition(name='response_message_level'))

    response_message = Quantity(
        type=str,
        shape=[],
        description='''
        Message outputted by the program formatting the data in the current format
        ''',
        a_legacy=LegacyDefinition(name='response_message'))


class section_soap_coefficients(MSection):
    '''
    Stores the soap coefficients for the pair of atoms given in
    soap_coefficients_atom_pair.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_soap_coefficients'))

    number_of_soap_coefficients = Quantity(
        type=int,
        shape=[],
        description='''
        number of soap coefficients
        ''',
        a_legacy=LegacyDefinition(name='number_of_soap_coefficients'))

    soap_coefficients_atom_pair = Quantity(
        type=str,
        shape=[],
        description='''
        Pair of atoms described in the current section
        ''',
        a_legacy=LegacyDefinition(name='soap_coefficients_atom_pair'))

    soap_coefficients = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_soap_coefficients'],
        description='''
        Compressed coefficient of the soap descriptor for the atom pair
        soap_coefficients_atom_pair
        ''',
        a_legacy=LegacyDefinition(name='soap_coefficients'))


class section_soap(MSection):
    '''
    Stores a soap descriptor for this configuration.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_soap'))

    soap_angular_basis_L = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        angular basis L
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_angular_basis_L'))

    soap_angular_basis_type = Quantity(
        type=str,
        shape=[],
        description='''
        angular basis type
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_angular_basis_type'))

    soap_kernel_adaptor = Quantity(
        type=str,
        shape=[],
        description='''
        kernel adaptor
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_kernel_adaptor'))

    soap_parameters_gid = Quantity(
        type=str,
        shape=[],
        description='''
        Unique checksum of all the soap parameters (all those with abstract type
        soap_parameter) with prefix psoap
        ''',
        a_legacy=LegacyDefinition(name='soap_parameters_gid'))

    soap_radial_basis_integration_steps = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        radial basis integration steps
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_radial_basis_integration_steps'))

    soap_radial_basis_mode = Quantity(
        type=str,
        shape=[],
        description='''
        radial basis mode
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_radial_basis_mode'))

    soap_radial_basis_n = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        radial basis N
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_radial_basis_n'))

    soap_radial_basis_sigma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        radial basis sigma
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_radial_basis_sigma'))

    soap_radial_basis_type = Quantity(
        type=str,
        shape=[],
        description='''
        radial basis type
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_radial_basis_type'))

    soap_radial_cutoff_center_weight = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        radial cutoff center weight
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_radial_cutoff_center_weight'))

    soap_radial_cutoff_rc_width = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        radial cutoff width
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_radial_cutoff_rc_width'))

    soap_radial_cutoff_rc = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        radial cutoff
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_radial_cutoff_rc'))

    soap_radial_cutoff_type = Quantity(
        type=str,
        shape=[],
        description='''
        radial cutoff type
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_radial_cutoff_type'))

    soap_spectrum_2l1_norm = Quantity(
        type=bool,
        shape=[],
        description='''
        2l1 norm spectrum
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_spectrum_2l1_norm'))

    soap_spectrum_global = Quantity(
        type=bool,
        shape=[],
        description='''
        global spectrum
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_spectrum_global'))

    soap_spectrum_gradients = Quantity(
        type=bool,
        shape=[],
        description='''
        gradients in specturm
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_spectrum_gradients'))

    soap_type_list = Quantity(
        type=str,
        shape=[],
        description='''
        Type list
        ''',
        categories=[soap_parameter],
        a_legacy=LegacyDefinition(name='soap_type_list'))

    section_soap_coefficients = SubSection(
        sub_section=SectionProxy('section_soap_coefficients'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_soap_coefficients'))


class section_topology(MSection):
    '''
    Section containing the definition of topology (connectivity among atoms in force
    fileds), force field, and constraints of a system.
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='section_topology'))

    atom_to_molecule = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_topology_atoms', 2],
        description='''
        Table mapping atom to molecules: the first column is the index of the molecule and
        the second column the index of the atom, signifying that the atom in the second
        column belongs to the molecule in the first column in the same row.
        ''',
        a_legacy=LegacyDefinition(name='atom_to_molecule'))

    molecule_to_molecule_type_map = Quantity(
        type=Reference(SectionProxy('section_molecule_type')),
        shape=['number_of_topology_molecules'],
        description='''
        Mapping from molecules to molecule types.
        ''',
        a_legacy=LegacyDefinition(name='molecule_to_molecule_type_map'))

    number_of_topology_atoms = Quantity(
        type=int,
        shape=[],
        description='''
        Number of atoms in the system described by this topology.
        ''',
        a_legacy=LegacyDefinition(name='number_of_topology_atoms'))

    number_of_topology_molecules = Quantity(
        type=int,
        shape=[],
        description='''
        Number of molecules in the system, as described by this topology.
        ''',
        a_legacy=LegacyDefinition(name='number_of_topology_molecules'))

    topology_force_field_name = Quantity(
        type=str,
        shape=[],
        description='''
        A unique string idenfiying the force field defined in this section. Strategies to
        define it are discussed in the
        [topology\\_force\\_field\\_name](https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-
        info/wikis/metainfo/topology-force-field-name).
        ''',
        a_legacy=LegacyDefinition(name='topology_force_field_name'))

    section_atom_type = SubSection(
        sub_section=SectionProxy('section_atom_type'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_atom_type'))

    section_constraint = SubSection(
        sub_section=SectionProxy('section_constraint'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_constraint'))

    section_interaction = SubSection(
        sub_section=SectionProxy('section_interaction'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_interaction'))

    section_molecule_type = SubSection(
        sub_section=SectionProxy('section_molecule_type'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_molecule_type'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    dft_plus_u_functional = Quantity(
        type=str,
        shape=[],
        description='''
        Type of DFT+U functional (such as DFT/DFT+U double-counting compensation). Valid
        names are described in the [dft\\_plus\\_u\\_functional wiki
        page](https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/dft-
        plus-u-functional).
        ''',
        a_legacy=LegacyDefinition(name='dft_plus_u_functional'))

    dft_plus_u_projection_type = Quantity(
        type=str,
        shape=[],
        description='''
        DFT+U: Type of orbitals used for projection in order to calculate occupation
        numbers. Valid names are described in the [dft\\_plus\\_u\\_projection\\_type wiki
        page](https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/dft-
        plus-u-projection-type).
        ''',
        a_legacy=LegacyDefinition(name='dft_plus_u_projection_type'))

    gw_bare_coulomb_cutofftype = Quantity(
        type=str,
        shape=[],
        description='''
        Cutoff type for the calculation of the bare Coulomb potential: none, 0d, 1d, 2d.
        See Rozzi et al., PRB 73, 205119 (2006)
        ''',
        a_legacy=LegacyDefinition(name='gw_bare_coulomb_cutofftype'))

    gw_bare_coulomb_gmax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / meter',
        description='''
        Maximum G for the pw basis for the Coulomb potential.
        ''',
        a_legacy=LegacyDefinition(name='gw_bare_coulomb_gmax'))

    gw_basis_set = Quantity(
        type=str,
        shape=[],
        description='''
        Auxillary basis set used for non-local operators: mixed - mixed basis set, Kotani
        and Schilfgaarde, Solid State Comm. 121, 461 (2002).
        ''',
        a_legacy=LegacyDefinition(name='gw_basis_set'))

    gw_core_treatment = Quantity(
        type=str,
        shape=[],
        description='''
        It specifies whether the core states are treated in the GW calculation: all - All
        electron calculation; val - Valence electron only calculation; vab - Core
        electrons are excluded from the mixed product basis; xal - All electron treatment
        of the exchange self-energy only
        ''',
        a_legacy=LegacyDefinition(name='gw_core_treatment'))

    gw_frequency_grid_type = Quantity(
        type=str,
        shape=[],
        description='''
        Frequency integration grid type for the correlational self energy: 'eqdis' -
        equidistant frequencies from 0 to freqmax; 'gaulag' - Gauss-Laguerre quadrature
        from 0 to infinity; 'gauleg' - Gauss-Legendre quadrature from 0 to freqmax;
        'gaule2' (default) - double Gauss-Legendre quadrature from 0 to freqmax and from
        freqmax to infinity.
        ''',
        a_legacy=LegacyDefinition(name='gw_frequency_grid_type'))

    gw_max_frequency = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Maximum frequency for the calculation of the self energy.
        ''',
        a_legacy=LegacyDefinition(name='gw_max_frequency'))

    gw_mixed_basis_gmax = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='1 / meter',
        description='''
        Cut-off parameter for the truncation of the expansion of the plane waves in the
        interstitial region.
        ''',
        a_legacy=LegacyDefinition(name='gw_mixed_basis_gmax'))

    gw_mixed_basis_lmax = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Maximum l value used for the radial functions within the muffin-tin.
        ''',
        a_legacy=LegacyDefinition(name='gw_mixed_basis_lmax'))

    gw_mixed_basis_tolerance = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Eigenvalue threshold below which the egenvectors are discarded in the construction
        of the radial basis set.
        ''',
        a_legacy=LegacyDefinition(name='gw_mixed_basis_tolerance'))

    gw_ngridq = Quantity(
        type=np.dtype(np.int32),
        shape=[3],
        description='''
        k/q-point grid size used in the GW calculation.
        ''',
        a_legacy=LegacyDefinition(name='gw_ngridq'))

    gw_frequency_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number referring to the frequency used in the calculation of the self energy.
        ''',
        a_legacy=LegacyDefinition(name='gw_frequency_number'))

    gw_frequency_values = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Values of the frequency used in the calculation of the self energy.
        ''',
        a_legacy=LegacyDefinition(name='gw_frequency_values'))

    gw_frequency_weights = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Weights of the frequency used in the calculation of the self energy.
        ''',
        a_legacy=LegacyDefinition(name='gw_frequency_weights'))

    gw_number_of_frequencies = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of frequency points used in the calculation of the self energy.
        ''',
        a_legacy=LegacyDefinition(name='gw_number_of_frequencies'))

    gw_polarizability_number_of_empty_states = Quantity(
        type=int,
        shape=[],
        description='''
        Number of empty states used to compute the polarizability P
        ''',
        a_legacy=LegacyDefinition(name='gw_polarizability_number_of_empty_states'))

    gw_qp_equation_treatment = Quantity(
        type=str,
        shape=[],
        description='''
        Methods to solve the quasi-particle equation: 'linearization', 'self-consistent'
        ''',
        a_legacy=LegacyDefinition(name='gw_qp_equation_treatment'))

    gw_screened_coulomb_volume_average = Quantity(
        type=str,
        shape=[],
        description='''
        Type of volume averaging for the dynamically screened Coulomb potential: isotropic
        - Simple averaging along a specified direction using only diagonal components of
        the dielectric tensor; anisotropic - Anisotropic screening by C. Freysoldt et al.,
        CPC 176, 1-13 (2007)
        ''',
        a_legacy=LegacyDefinition(name='gw_screened_coulomb_volume_average'))

    gw_screened_Coulomb = Quantity(
        type=str,
        shape=[],
        description='''
        Model used to calculate the dinamically-screened Coulomb potential: 'rpa' - Full-
        frequency random-phase approximation; 'ppm' - Godby-Needs plasmon-pole model Godby
        and Needs, Phys. Rev. Lett. 62, 1169 (1989); 'ppm_hl' - Hybertsen and Louie, Phys.
        Rev. B 34, 5390 (1986); 'ppm_lh' - von der Linden and P. Horsh, Phys. Rev. B 37,
        8351 (1988); 'ppm_fe' - Farid and Engel, Phys. Rev. B 47,15931 (1993); 'cdm' -
        Contour deformation method, Phys. Rev. B 67, 155208 (2003).)
        ''',
        a_legacy=LegacyDefinition(name='gw_screened_Coulomb'))

    gw_self_energy_c_analytical_continuation = Quantity(
        type=str,
        shape=[],
        description='''
        Models for the correlation self-energy analytical continuation: 'pade' -  Pade's
        approximant (by H. J. Vidberg and J. W. Serence, J. Low Temp. Phys. 29, 179
        (1977)); 'mpf' -  Multi-Pole Fitting (by H. N Rojas, R. W. Godby and R. J. Needs,
        Phys. Rev. Lett. 74, 1827 (1995)); 'cd' - contour deformation; 'ra' - real axis
        ''',
        a_legacy=LegacyDefinition(name='gw_self_energy_c_analytical_continuation'))

    gw_self_energy_c_number_of_empty_states = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of empty states to be used to calculate the correlation self energy.
        ''',
        a_legacy=LegacyDefinition(name='gw_self_energy_c_number_of_empty_states'))

    gw_self_energy_c_number_of_poles = Quantity(
        type=int,
        shape=[],
        description='''
        Number of poles used in the analytical continuation.
        ''',
        a_legacy=LegacyDefinition(name='gw_self_energy_c_number_of_poles'))

    gw_self_energy_singularity_treatment = Quantity(
        type=str,
        shape=[],
        description='''
        Treatment of the integrable singular terms in the calculation of the self energy.
        Values: 'mpb' - Auxiliary function method by S. Massidda, M. Posternak, and A.
        Baldereschi, PRB 48, 5058 (1993); 'crg' - Auxiliary function method by P. Carrier,
        S. Rohra, and A. Goerling, PRB 75, 205126 (2007).
        ''',
        a_legacy=LegacyDefinition(name='gw_self_energy_singularity_treatment'))

    gw_starting_point = Quantity(
        type=str,
        shape=[],
        description='''
        Exchange-correlation functional of the ground-state calculation. See XC_functional
        list at https://gitlab.mpcdf.mpg.de/nomad-lab/nomad-meta-info/wikis/metainfo/XC-
        functional
        ''',
        a_legacy=LegacyDefinition(name='gw_starting_point'))

    gw_type_test = Quantity(
        type=str,
        shape=[],
        description='''
        GW methodology: exciting test variable
        ''',
        a_legacy=LegacyDefinition(name='gw_type_test'))

    gw_type = Quantity(
        type=str,
        shape=[],
        description='''
        GW methodology: G0W0; ev-scGW: (eigenvalues self-consistent GW) – Phys.Rev.B 34,
        5390 (1986); qp-scGW: (quasi-particle self-consistent GW) – Phys. Rev. Lett. 96,
        226402 (2006)  scGW0: (self-consistent G with fixed W0) – Phys.Rev.B 54, 8411
        (1996); scG0W: (self-consistent W with fixed G0); scGW: (self-consistent GW) –
        Phys. Rev. B 88, 075105 (2013)
        ''',
        a_legacy=LegacyDefinition(name='gw_type'))

    method_to_topology_ref = Quantity(
        type=Reference(SectionProxy('section_topology')),
        shape=[],
        description='''
        Reference to the topology and force fields to be used.
        ''',
        a_legacy=LegacyDefinition(name='method_to_topology_ref'))

    section_dft_plus_u_orbital = SubSection(
        sub_section=SectionProxy('section_dft_plus_u_orbital'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_dft_plus_u_orbital'))

    section_method_basis_set = SubSection(
        sub_section=SectionProxy('section_method_basis_set'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_method_basis_set'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    energy_C_mGGA = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Component of the correlation (C) energy at the GGA (or MetaGGA) level using the
        self-consistent density of the target XC functional (full unscaled value, i.e.,
        not scaled due to exact-exchange mixing).
        ''',
        categories=[public.energy_component, public.energy_value, public.energy_type_C],
        a_legacy=LegacyDefinition(name='energy_C_mGGA'))

    energy_reference_fermi = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels'],
        unit='joule',
        description='''
        Fermi energy (separates occupied from unoccupied single-particle states in metals)
        ''',
        categories=[public.energy_type_reference, public.energy_value],
        a_legacy=LegacyDefinition(name='energy_reference_fermi'))

    energy_reference_highest_occupied = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels'],
        unit='joule',
        description='''
        Highest occupied single-particle state energy (in insulators or HOMO energy in
        finite systems)
        ''',
        categories=[public.energy_type_reference, public.energy_value],
        a_legacy=LegacyDefinition(name='energy_reference_highest_occupied'))

    energy_reference_lowest_unoccupied = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels'],
        unit='joule',
        description='''
        Lowest unoccupied single-particle state energy (in insulators or LUMO energy in
        finite systems)
        ''',
        categories=[public.energy_type_reference, public.energy_value],
        a_legacy=LegacyDefinition(name='energy_reference_lowest_unoccupied'))

    energy_X_mGGA_scaled = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Component of the exchange (X) energy at the GGA (or MetaGGA) level, using the self
        consistent density of the target functional, scaled accordingly to the mixing
        parameter.
        ''',
        categories=[public.energy_component, public.energy_value],
        a_legacy=LegacyDefinition(name='energy_X_mGGA_scaled'))

    energy_X_mGGA = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Component of the exchange (X) energy at the GGA (or MetaGGA) level using the self
        consistent density of the target functional (full unscaled value, i.e., not scaled
        due to exact-exchange mixing).
        ''',
        categories=[public.energy_type_X, public.energy_component, public.energy_value],
        a_legacy=LegacyDefinition(name='energy_X_mGGA'))

    gw_fermi_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        GW Fermi energy
        ''',
        a_legacy=LegacyDefinition(name='gw_fermi_energy'))

    gw_fundamental_gap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        GW fundamental band gap
        ''',
        a_legacy=LegacyDefinition(name='gw_fundamental_gap'))

    gw_optical_gap = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        GW optical band gap
        ''',
        a_legacy=LegacyDefinition(name='gw_optical_gap'))

    gw_self_energy_c = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels', 'number_of_eigenvalues_kpoints', 'number_of_eigenvalues'],
        unit='joule',
        description='''
        Diagonal matrix elements of the correlation self-energy
        ''',
        a_legacy=LegacyDefinition(name='gw_self_energy_c'))

    gw_self_energy_x = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels', 'number_of_eigenvalues_kpoints', 'number_of_eigenvalues'],
        unit='joule',
        description='''
        Diagonal matrix elements of the exchange self-energy
        ''',
        a_legacy=LegacyDefinition(name='gw_self_energy_x'))

    gw_xc_potential = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels', 'number_of_eigenvalues_kpoints', 'number_of_eigenvalues'],
        unit='joule',
        description='''
        Diagonal matrix elements of the exchange-correlation potential
        ''',
        a_legacy=LegacyDefinition(name='gw_xc_potential'))

    section_excited_states = SubSection(
        sub_section=SectionProxy('section_excited_states'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_excited_states'))


class section_scf_iteration(public.section_scf_iteration):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_scf_iteration'))

    energy_reference_fermi_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels'],
        unit='joule',
        description='''
        Fermi energy (separates occupied from unoccupied single-particle states in metals)
        during the self-consistent field (SCF) iterations.
        ''',
        categories=[public.energy_type_reference, public.energy_value],
        a_legacy=LegacyDefinition(name='energy_reference_fermi_iteration'))

    energy_reference_highest_occupied_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels'],
        unit='joule',
        description='''
        Highest occupied single-particle state energy (in insulators or HOMO energy in
        finite systems) during the self-consistent field (SCF) iterations.
        ''',
        categories=[public.energy_type_reference, public.energy_value],
        a_legacy=LegacyDefinition(name='energy_reference_highest_occupied_iteration'))

    energy_reference_lowest_unoccupied_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels'],
        unit='joule',
        description='''
        Lowest unoccupied single-particle state energy (in insulators or LUMO energy in
        finite systems) during the self-consistent field (SCF) iterations.
        ''',
        categories=[public.energy_type_reference, public.energy_value],
        a_legacy=LegacyDefinition(name='energy_reference_lowest_unoccupied_iteration'))


class section_eigenvalues(public.section_eigenvalues):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_eigenvalues'))

    gw_qp_linearization_prefactor = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels', 'number_of_eigenvalues_kpoints', 'number_of_eigenvalues'],
        description='''
        Linearization prefactor
        ''',
        a_legacy=LegacyDefinition(name='gw_qp_linearization_prefactor'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    number_of_electrons = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_spin_channels'],
        description='''
        Number of electrons in system
        ''',
        categories=[public.configuration_core],
        a_legacy=LegacyDefinition(name='number_of_electrons'))

    topology_ref = Quantity(
        type=Reference(SectionProxy('section_topology')),
        shape=[],
        description='''
        Reference to the topology used for this system; if not given, the trivial topology
        should be assumed.
        ''',
        a_legacy=LegacyDefinition(name='topology_ref'))

    is_representative = Quantity(
        type=bool,
        shape=[],
        description='''
        Most systems in a run are only minor variations of each other. Systems marked
        representative where chosen to be representative for all systems in the run.
        ''',
        a_legacy=LegacyDefinition(name='is_representative'))

    section_soap = SubSection(
        sub_section=SectionProxy('section_soap'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_soap'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    section_topology = SubSection(
        sub_section=SectionProxy('section_topology'),
        repeats=True,
        a_legacy=LegacyDefinition(name='section_topology'))


m_package.__init_metainfo__()



# Copyright 2016-2018 Mikkel Strange, Fawzi Mohamed
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from ase import Atoms
from ase.lattice.triclinic import Triclinic as TC
from ase.build import bulk


class UnitCell:
    def __init__(self, a, b, c, origin=None):
        self.cell = [a, b, c]


class Triclinic:
    def __init__(self, a, b, c, alpha, beta, gamma):
        self.cell = TC(symbol='X', latticeconstant=(a, b, c, alpha,
                                                    beta, gamma)).get_cell()
    def primitiveVectors(self):
        return self.cell

class FaceCenteredCubic:
    def __init__(self, a):
        self.cell = bulk('X', crystalstructure='fcc', a=a).get_cell()

    def primitiveVectors(self):
        return self.cell


class BodyCenteredCubic:
    def __init__(self, a):
        self.cell = bulk('X', crystalstructure='bcc', a=a).get_cell()

    def primitiveVectors(self):
        return self.cell


class BulkConfiguration:
    def __init__(self, bravais_lattice, elements, cartesian_coordinates=None,
                 fractional_coordinates=None, ghost_atoms=None,
                 velocities=None, tag_data=None, fast_init=False):
        if elements is None:
            symbols = None
        else:
            symbols = [e.symbol() for e in elements]
        if cartesian_coordinates is not None:
            positions = cartesian_coordinates
            scale_atoms = False
        elif fractional_coordinates is not None:
            positions = fractional_coordinates
            scale_atoms = True
        else:
            positions = None
            scale_atoms = False
        pbc = True
        atoms = Atoms(symbols, positions, pbc=pbc)
        atoms.set_cell(bravais_lattice.cell, scale_atoms=scale_atoms)
        atoms.ghost_atoms = ghost_atoms
        self.atoms = atoms


class MoleculeConfiguration:
    def __init__(self, elements=None, cartesian_coordinates=None,
                 fractional_coordinates=None, ghost_atoms=None,
                 velocities=None, tag_data=None, fast_init=False):
        if elements is None:
            symbols = None
        else:
            symbols = [e.symbol() for e in elements]
        if cartesian_coordinates is not None:
            positions = cartesian_coordinates
        else:
            positions = None
        pbc = False
        atoms = Atoms(symbols, positions, pbc=pbc)
        atoms.ghost_atoms = ghost_atoms
        self.atoms = atoms


things = {'UnitCell': UnitCell,
          'Triclinic': Triclinic,
          'FaceCenteredCubic': FaceCenteredCubic,
          'BulkConfiguration': BulkConfiguration,
          'MoleculeConfiguration': MoleculeConfiguration}

conf_types = ['MoleculeConfiguration',
              'BulkConfiguration']

if __name__ == '__main__':
    t = Triclinic(4.0, 4.0, 4.0, 90.0, 90.0, 90.0)
    print(t.primitiveVectors())



#!/usr/bin/env python
"""
Normalizes i.e. checks consistency, completness, absence of duplicates, and reformats meta_info data
"""
from builtins import str
import sys, os, os.path, datetime
basePath = os.path.realpath(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
if not basePath in sys.path:
    sys.path.append(basePath)
from nomadcore.local_meta_info import InfoKindEnv, loadJsonFile, InfoKindEl
from nomadcore.json_support import jsonIndentD, jsonCompactD
from nomadcore.compact_sha import sha512

def normalizeFile(path, checkOnly, force, addGid, addSubGids, extraArgsHandling, s = sys.stderr.write):
    """Normalizes the meta info file at the given path"""
    env, warnings = loadJsonFile(path, extraArgsHandling = extraArgsHandling)
    hasErrors = False
    ovr = warnings.get("overrides")
    relpath = os.path.relpath(path)
    if ovr:
        hasErrors = True
        s("Error: loading {0!r}, found duplicates:\n".format(relpath))
        for old, new in ovr:
            jsonIndentD(old, s, extraIndent = 4)
            s(" => ")
            jsonIndentD(new, s, extraIndent = 4)
            s("\n")
    dup = warnings.get("duplicates")
    if dup:
        hasErrors = True
        s("Error: loading {0!r},\n    the depenendencies already define kinds with the following names:\n    ".format(relpath))
        s(str(dup))
        s("\n")
    hid = warnings.get("hidden")
    if hid:
        hasErrors = True
        s("Error: loading {0!r},\n    the depenendencies already define different kinds with the same names:\n    ".format(
                relpath))
        s(str([x.toDict() for x in hid]))
        s("\n")
    res = {"path":path, "env": env, "hasErrors": hasErrors, "hasChanges": False}
    try:
        env1 = InfoKindEnv(name = env.name, description = env.description, infoKinds=list(env.infoKinds.values()), path=env.path, uri=env.uri, deps=env.deps)
        env1.calcGids()
        validGids = True
    except:
        validGids = False
        hasErrors = True
        s("Error: parsing {0!r} failed to re-generate gids: {1}\n".format(relpath, sys.exc_info()[1]))
    if (hasErrors and not force):
        return res
    if env.gids and validGids:
        changes = []
        noDepNames = env.noDepNames()
        for k, v in env.gids.items():
            if k in noDepNames and k in env1.gids and env1.gids.get(k) != v:
                changes.append(k)
        if changes:
            s("Info: {0!r} regenerating gids changed for: ".format(relpath))
            jsonCompactD(changes, s)
            s("\n")
    if validGids:
        env = env1
    oldSha = None
    with open(path) as f:
        sha = sha512()
        while True:
            block = f.read(8*1024)
            if not block:
                break
            sha.update(block)
        oldSha = sha.b64digest()
    newS = sha512()
    try:
        env.serialize(newS.update, subGids = addSubGids, selfGid = addGid)
        newSha = newS.b64digest()
    except:
        hasErrors = True
        res["hasErrors"] = hasErrors
        s("Error: serialization of {0!r} failed: {1}".format(relpath, sys.exc_info()[1]))
        return res
    backup = None
    if (oldSha == newSha):
        s(repr(path))
        s(": OK, no change\n")
    else:
        backupBase = path + "-" + datetime.date.today().isoformat()
        extra = ""
        n = 0
        while os.path.exists(backupBase + extra + ".bk"):
            n += 1
            extra = "-" + str(n)
        backup = backupBase + extra + ".bk"
        if not checkOnly:
            os.rename(path, backup)
            with open(path, "wb") as f:
                env.serialize(f.write, subGids = addSubGids, selfGid = addGid)
        s(repr(path))
        s(": %s, has changes, backup in " % ("OK" if not hasErrors else "tying to fix ERRORS"))
        s(os.path.basename(backup))
        s("\n")
        res["hasChanges"] = True
        res["backup"] = backup
    return res

def normalizePath(path, checkOnly, force, addGid, addSubGids, extraArgsHandling, s = sys.stderr.write):
    """Normalizes all the nomadmetainfo.json files at the given path"""
    if (os.path.isfile(path)):
        return normalizeFile(path, checkOnly=checkOnly, force=force, addGid=addGid, addSubGids=addSubGids, extraArgsHandling=extraArgsHandling, s=s)
    envNames = {}
    allNames = set()
    hasErrors = False
    hasChanges = False
    clashes = set()
    for dirpath, dirnames, filenames in os.walk(path):
        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            if not filename.endswith(".nomadmetainfo.json"):
                if not filename.endswith(".bk") and not filename.startswith("."):
                    s("skipping ")
                    s(filepath)
                    s("\n")
                continue
            fileRes = normalizeFile(filepath, checkOnly=checkOnly, force=force,
                    addGid=addGid, addSubGids=addSubGids, extraArgsHandling=extraArgsHandling, s=s)
            hasChanges = hasChanges or fileRes.get("hasChanges", False)
            hasErrors = hasErrors or fileRes.get("hasErrors", False)
            newNames = fileRes.get("env").noDepNames()
            envNames[filepath] = newNames
            newClashes = allNames.intersection(newNames)
            allNames.update(newNames)
            clashes.update(newClashes)
    detailedClashes = {}
    for filePath, names in envNames.items():
        currentClashes = names.intersection(clashes)
        if currentClashes:
            detailedClashes[filePath]=currentClashes
    if clashes:
        hasErrors = True
        s(repr(path))
        s(", detected clashes:\n")
        for filePath, fileClashes in detailedClashes.items():
            s("    ")
            s(filePath)
            s(":\n        ")
            jsonCompactD(list(fileClashes), s)
            s("\n")
    return {"path": path, "hasErrors": hasErrors, "hasChanges": hasChanges}

def normalizePaths(paths, checkOnly, force, addGid, addSubGids, extraArgsHandling, s = sys.stderr.write):
    """Normalizes all the given paths"""
    hasErrors = False
    hasChanges = False
    messages = []
    for path in paths:
        messages.append(path)
        res = normalizePath(path, checkOnly=checkOnly, force=force,
                addGid=addGid, addSubGids=addSubGids, extraArgsHandling=extraArgsHandling, s=s)
        errors = res.get("hasErrors", True)
        hasErrors = hasErrors or errors
        if not errors:
            messages.append(": OK")
        else:
            messages.append(": ERROR")
        changes = res.get("hasChanges", False)
        hasChanges = hasChanges or changes
        if changes:
            messages.append(", with changes\n")
        else:
            messages.append(", no changes\n")
    s("\n")
    for m in messages:
        s(m)
    s("\n")
    if not hasErrors:
        s("OK")
    else:
        s("ERROR")
    if hasChanges:
        s(", with changes\n")
        if checkOnly:
            s("    (run performed with --check-only, no changes were committed to the filesystem)\n")
    else:
        s(", no changes\n")
    return {"paths": paths, "hasErrors": hasErrors, "hasChanges": hasChanges}

if __name__ == "__main__":
    defaultDir = os.path.normpath(os.path.join(basePath, "../../../nomad-meta-info/meta_info/nomad_meta_info"))
    usage = """usage: {command} [--check-only] [--force] [--add-gid] [--add-sub-gids] [--keep-extra-args] [--remove-extra-args] [--error-if-extra-args] [file/or/dir ...]

    normalizes the InfoKinds file/or/dir (that defalts to {dir})
""".format(command=os.path.basename(sys.argv[0]), dir=defaultDir)

    checkOnly = False
    force = False
    addGid = False
    addSubGids = False
    paths = []
    extraArgsHandling = InfoKindEl.ADD_EXTRA_ARGS

    for arg in sys.argv[1:]:
        if arg == "--check-only":
            checkOnly = True
        elif arg == "--force":
            force = True
        elif arg == "--add-gid":
            addGid = True
        elif arg == "--add-sub-gids":
            addSubGids = True
        elif arg == "--add-extra-args":
            extraArgsHandling = InfoKindEl.ADD_EXTRA_ARGS
        elif arg == "--remove-extra-args":
            extraArgsHandling = InfoKindEl.IGNORE_EXTRA_ARGS
        elif arg == "--error-if-extra-args":
            extraArgsHandling = InfoKindEl.RAISE_IF_EXTRA_ARGS
        elif arg == "--help":
            sys.stdout.write(usage)
            sys.exit(0)
        elif os.path.exists(arg):
            paths.append(arg)
        else:
            sys.stdout.write("Error: invalid argument ")
            sys.stdout.write(repr(arg))
            sys.stdout.write("\n")
            sys.stdout.write(usage)
            sys.exit(1)
    if not paths:
        paths.append(defaultDir)
    normalizePaths(paths, checkOnly=checkOnly, force=force, addGid=addGid, addSubGids=addSubGids, extraArgsHandling=extraArgsHandling, s = sys.stderr.write)



# Copyright 2016-2018 Arvid Conrad Ihrig, Aliaksei Mazheika
#                     Fritz-Haber-Institut der Max-Planck-Gesellschaft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys, os, os.path
baseDir = os.path.dirname(os.path.abspath(__file__))
commonDir = os.path.normpath(os.path.join(baseDir,"../../../../python-common/common/python"))

if not commonDir in sys.path:
    sys.path.insert(0, commonDir)



import numpy as np
import os
import logging
from ase import data as asedata

from .LammpsCommon import converter

from nomad.parsing.text_parser import Quantity, UnstructuredTextFileParser
from nomad.datamodel.metainfo.public import section_run, section_sampling_method,\
    section_system, section_single_configuration_calculation, section_energy_contribution,\
    Workflow, MolecularDynamics
from nomad.datamodel.metainfo.common import section_topology, section_interaction


class Parser:
    def __init__(self, mainfile, logger):
        self._mainfile = mainfile
        self._results = None
        self._quantities = None
        self.logger = logger

    def __getitem__(self, key):
        val = self.results.get(key, None)
        return val

    def get(self, key, default=None):
        val = self[key]
        if not val:
            return default
        return val

    @property
    def maindir(self):
        return os.path.dirname(self._mainfile)

    @property
    def mainfile(self):
        return self._mainfile

    @mainfile.setter
    def mainfile(self, value):
        self._mainfile = value
        self._results = None

    @property
    def results(self):
        if self._results is None:
            self.parse()
        if self._results is None:
            self._results = {}

        return self._results

    def parse(self):
        pass


class DataParser(Parser):
    def __init__(self, mainfile, logger):
        super().__init__(mainfile, logger)
        self._headers = [
            'atoms', 'bonds', 'angles', 'dihedrals', 'impropers', 'atom types', 'bond types',
            'angle types', 'dihedral types', 'improper types', 'extra bond per atom',
            'extra/bond/per/atom', 'extra angle per atom', 'extra/angle/per/atom',
            'extra dihedral per atom', 'extra/dihedral/per/atom', 'extra improper per atom',
            'extra/improper/per/atom', 'extra special per atom', 'extra/special/per/atom',
            'ellipsoids', 'lines', 'triangles', 'bodies', 'xlo xhi', 'ylo yhi', 'zlo zhi',
            'xy xz yz']
        self._sections = [
            'Atoms', 'Velocities', 'Masses', 'Ellipsoids', 'Lines', 'Triangles', 'Bodies',
            'Bonds', 'Angles', 'Dihedrals', 'Impropers', 'Pair Coeffs', 'PairIJ Coeffs',
            'Bond Coeffs', 'Angle Coeffs', 'Dihedral Coeffs', 'Improper Coeffs',
            'BondBond Coeffs', 'BondAngle Coeffs', 'MiddleBondTorsion Coeffs',
            'EndBondTorsion Coeffs', 'AngleTorsion Coeffs', 'AngleAngleTorsion Coeffs',
            'BondBond13 Coeffs', 'AngleAngle Coeffs']
        self._interactions = [
            section for section in self._sections if section.endswith('Coeffs')]

    def parse(self):
        if self.mainfile is None:
            return

        if not os.path.isfile(self.mainfile):
            return

        if self._quantities is None:
            self._quantities = []
            for header in self._headers:
                self._quantities.append(
                    Quantity(header, r'\s*([\+\-eE\d\. ]+)\s*%s\s*\n' % header, comment='#'))

            def get_section_value(val):
                val = val.split('\n')
                name = None

                if val[0][0] == '#':
                    name = val[0][1:].strip()
                    val = val[1:]

                value = []
                for i in range(len(val)):
                    v = val[i].split('#')[0].split()
                    if not v:
                        continue

                    try:
                        value.append(np.array(v, dtype=float))
                    except Exception:
                        break

                return name, np.array(value)

            for section in self._sections:
                self._quantities.append(
                    Quantity(
                        section, r'\s*%s\s*(#*\s*[\s\S]*?\n)\n*([\deE\-\+\.\s]+)\n' % section,
                        str_operation=get_section_value))

        parser = UnstructuredTextFileParser(self.mainfile, self._quantities)

        self._results = {key: parser[key] for key in parser.keys() if parser[key]}

    def get_interactions(self):
        styles_coeffs = []
        for interaction in self._interactions:
            coeffs = self.get(interaction, None)
            if coeffs is None:
                continue
            if isinstance(coeffs, tuple):
                coeffs = [coeffs]

            styles_coeffs += coeffs

        return styles_coeffs


class TrajParser(Parser):
    def __init__(self, mainfile, logger):
        super().__init__(mainfile, logger)
        self._masses = None
        self._reference_masses = dict(
            masses=np.array(asedata.atomic_masses), symbols=asedata.chemical_symbols)
        self._chemical_symbols = None

    def parse(self):
        if self.mainfile is None:
            return

        if not os.path.isfile(self.mainfile):
            return

        if self._quantities is None:
            def get_pbc_cell(val):
                val = val.split()
                pbc = [v == 'pp' for v in val[:3]]

                cell = np.zeros((3, 3))
                for i in range(3):
                    cell[i][i] = float(val[i * 2 + 4]) - float(val[i * 2 + 3])

                return pbc, cell

            def get_atoms_info(val):
                val = val.split('\n')
                keys = val[0].split()
                values = np.array([v.split() for v in val[1:] if v], dtype=float)
                values = values[values[:, 0].argsort()].T
                return {keys[i]: values[i] for i in range(len(keys))}

            self._quantities = [
                Quantity(
                    'time_step', r'\s*ITEM:\s*TIMESTEP\s*\n\s*(\d+)\s*\n', comment='#'),
                Quantity(
                    'n_atoms', r'\s*ITEM:\s*NUMBER OF ATOMS\s*\n\s*(\d+)\s*\n', comment='#'),
                Quantity(
                    'pbc_cell', r'\s*ITEM: BOX BOUNDS\s*([\s\w]+)([\+\-\d\.eE\s]+)\n',
                    str_operation=get_pbc_cell, comment='#'),
                Quantity(
                    'atoms_info', r's*ITEM:\s*ATOMS\s*([ \w]+\n)*?([\+\-eE\d\.\n ]+)\n*I*',
                    str_operation=get_atoms_info, comment='#')
            ]

        parser = UnstructuredTextFileParser(self.mainfile, self._quantities)

        self._results = {key: parser[key] for key in parser.keys() if parser[key]}

    def with_trajectory(self):
        if self._results is None:
            return False

        return self._results.get('atoms_info', None) is not None

    @property
    def masses(self):
        return self._masses

    @masses.setter
    def masses(self, val):
        self._masses = val
        if self._masses is None:
            return

        self._masses = val
        if self._chemical_symbols is None:
            masses = self._masses[0][1]
            self._chemical_symbols = {}
            for i in range(len(masses)):
                symbol_idx = np.argmin(abs(self._reference_masses['masses'] - masses[i][1]))
                self._chemical_symbols[masses[i][0]] = self._reference_masses['symbols'][symbol_idx]

    def get_atom_labels(self, idx):
        atoms_info = self.results.get('atoms_info')[idx]
        atoms_type = atoms_info.get('type', None)
        if atoms_type is None:
            return

        if self._chemical_symbols is None:
            return

        atom_labels = [self._chemical_symbols[atype] for atype in atoms_type]

        return atom_labels

    def get_positions(self, idx):
        atoms_info = self.results.get('atoms_info')[idx]

        cell = self.results.get('pbc_cell')[idx][1]

        if 'xs' in atoms_info and 'ys' in atoms_info and 'zs' in atoms_info:
            positions = np.array([atoms_info['xs'], atoms_info['ys'], atoms_info['zs']]).T

            positions = positions * np.linalg.norm(cell, axis=1) + np.amin(cell, axis=1)

        else:
            positions = np.array([atoms_info['x'], atoms_info['y'], atoms_info['z']]).T

            if 'ix' in atoms_info and 'iy' in atoms_info and 'iz' in atoms_info:
                positions_img = np.array([
                    atoms_info['ix'], atoms_info['iy'], atoms_info['iz']]).T

                positions += positions_img * np.linalg.norm(cell, axis=1)

        return positions

    def get_velocities(self, idx):
        atoms_info = self.results.get('atoms_info')[idx]

        if 'vx' not in atoms_info or 'vy' not in atoms_info or 'vz' not in atoms_info:
            return

        return np.array([atoms_info['vx'], atoms_info['vy'], atoms_info['vz']]).T

    def get_forces(self, idx):
        atoms_info = self.results.get('atoms_info')[idx]

        if 'fx' not in atoms_info or 'fy' not in atoms_info or 'fz' not in atoms_info:
            return

        return np.array([atoms_info['fx'], atoms_info['fy'], atoms_info['fz']]).T


class LogParser(Parser):
    def __init__(self, mainfile, logger):
        super().__init__(mainfile, logger)
        self.logger = logger
        self._thermo_data = None
        self._commands = [
            'angle_coeff', 'angle_style', 'atom_modify', 'atom_style', 'balance',
            'bond_coeff', 'bond_style', 'bond_write', 'boundary', 'change_box', 'clear',
            'comm_modify', 'comm_style', 'compute', 'compute_modify', 'create_atoms',
            'create_bonds', 'create_box', 'delete_bonds', 'dielectric', 'dihedral_coeff',
            'dihedral_style', 'dimension', 'displace_atoms', 'dump', 'dump_modify',
            'dynamical_matrix', 'echo', 'fix', 'fix_modify', 'group', 'group2ndx',
            'ndx2group', 'hyper', 'if', 'improper_coeff', 'improper_style', 'include',
            'info', 'jump', 'kim_init', 'kim_interactions', 'kim_query', 'kim_param',
            'kim_property', 'kspace_modify', 'kspace_style', 'label', 'lattice', 'log',
            'mass', 'message', 'min_modify', 'min_style', 'minimize', 'minimize/kk',
            'molecule', 'neb', 'neb/spin', 'neigh_modify', 'neighbor', 'newton', 'next',
            'package', 'pair_coeff', 'pair_modify', 'pair_style', 'pair_write',
            'partition', 'prd', 'print', 'processors', 'quit', 'read_data', 'read_dump',
            'read_restart', 'region', 'replicate', 'rerun', 'reset_atom_ids',
            'reset_mol_ids', 'reset_timestep', 'restart', 'run', 'run_style', 'server',
            'set', 'shell', 'special_bonds', 'suffix', 'tad', 'temper/grem', 'temper/npt',
            'thermo', 'thermo_modify', 'thermo_style', 'third_order', 'timer', 'timestep',
            'uncompute', 'undump', 'unfix', 'units', 'variable', 'velocity', 'write_coeff',
            'write_data', 'write_dump', 'write_restart']
        self._interactions = [
            'atom', 'pair', 'bond', 'angle', 'dihedral', 'improper', 'kspace']

    def get_thermodynamic_data(self):
        def str_operation(val):
            res = {}
            if val.count('Step') > 1:
                val = val.replace('-', '').replace('=', '').replace('(sec)', '').split()
                val = [v.strip() for v in val]

                for i in range(len(val)):
                    if val[i][0].isalpha():
                        res.setdefault(val[i], [])
                        res[val[i]].append(float(val[i + 1]))

            else:
                val = val.split('\n')
                keys = [v.strip() for v in val[0].split()]
                val = np.array([v.split() for v in val[1:] if v], dtype=float).T

                res = {key: [] for key in keys}
                for i in range(len(keys)):
                    res[keys[i]] = val[i]

            return res

        pattern = r'\s*\-*(\s*Step\s*[\-\s\w\.\=\(\)]*[ \-\.\d\n]+)Loop'

        parser = UnstructuredTextFileParser(
            self.mainfile, [Quantity('thermo_data', pattern, str_operation=str_operation)])

        thermo_data = parser['thermo_data']

        self._thermo_data = list(thermo_data)[0] if thermo_data is not None else thermo_data

        return self._thermo_data

    def get_traj_files(self):
        dump = self.results.get('dump', None)
        if dump is None:
            self.logger.warn(
                'Trajectory not specified in %s, will scan directory' % self.maindir)
            traj_files = os.listdir(self.maindir)
            traj_files = [f for f in traj_files if f.endswith('trj')]

        else:
            traj_files = []
            if type(dump[0]) in [str, int]:
                dump = [dump]
            traj_files = [d[4] for d in dump]

        return [os.path.join(self.maindir, f) for f in traj_files]

    def get_data_files(self):
        read_data = self.results.get('read_data', None)
        if read_data is None:
            self.logger.warn(
                'Data file not specified in %s, will scan directory' % self.maindir)
            data_files = os.listdir(self.maindir)
            data_files = [f for f in data_files if f.endswith('data') or f.startswith('data')]

        else:
            data_files = read_data

        return [os.path.join(self.maindir, f) for f in data_files]

    def get_pbc(self):
        pbc = self.results.get('boundary', ['p', 'p', 'p'])
        return [v == 'p' for v in pbc]

    def get_program_version(self):
        version = self.results.get('program_version', [''])[0]
        return ' '.join(version)

    def get_sampling_method(self):
        fix_style = self.get('fix', [[''] * 3])[0][2]

        sampling_method = 'langevin_dynamics' if 'langevin' in fix_style else 'molecular_dynamics'
        return sampling_method, fix_style

    def get_thermostat_settings(self):
        fix = self.get('fix', [None])[0]
        if fix is None:
            return {}

        try:
            fix_style = fix[2]
        except IndexError:
            return {}

        res = dict()
        if fix_style.lower() == 'nvt':
            try:
                res['target_T'] = float(fix[5])
                res['thermostat_tau'] = float(fix[6])
            except Exception:
                pass

        elif fix_style.lower() == 'npt':
            try:
                res['target_T'] = float(fix[5])
                res['thermostat_tau'] = float(fix[6])
                res['target_P'] = float(fix[9])
                res['barostat_tau'] = float(fix[10])
            except Exception:
                pass

        elif fix_style.lower() == 'nph':
            try:
                res['target_P'] = float(fix[5])
                res['barostat_tau'] = float(fix[6])
            except Exception:
                pass

        elif fix_style.lower() == 'langevin':
            try:
                res['target_T'] = float(fix[4])
                res['langevin_gamma'] = float(fix[5])
            except Exception:
                pass

        else:
            self.logger.warn('Fix style %s not supported' % fix_style)

        return res

    def get_interactions(self):
        styles_coeffs = []
        for interaction in self._interactions:
            styles = self.get('%s_style' % interaction, None)
            if styles is None:
                continue

            if isinstance(styles[0], str):
                styles = [styles]

            for i in range(len(styles)):
                if interaction == 'kspace':
                    coeff = [float(c) for c in styles[i][1:]]
                    style = styles[i][0]

                else:
                    coeff = self.get("%s_coeff" % interaction)
                    style = ' '.join([str(si) for si in styles[i]])

                styles_coeffs.append((style.strip(), coeff))

        return styles_coeffs

    def finished_normally(self):
        return self.get('finished', None) is not None

    def with_thermodynamics(self):
        return self._thermo_data is not None

    def parse(self):
        if self.mainfile is None:
            return

        if not os.path.isfile(self.mainfile):
            return

        if self._quantities is None:
            def str_op(val):
                val = val.split('#')[0]
                val = val.replace('&\n', ' ').split()
                val = val if len(val) > 1 else val[0]
                return val

            self._quantities = [
                Quantity(
                    name, r'\n\s*%s\s+([\w\. \/\#\-]+)(\&\n[\w\. \/\#\-]*)*' % name,
                    str_operation=str_op, comment='#') for name in self._commands]

            self._quantities.append(
                Quantity('program_version', r'\s*LAMMPS\s*\(([\w ]+)\)\n', dtype=str)
            )

            self._quantities.append(
                Quantity('finished', r'\s*Dangerous builds\s*=\s*(\d+)')
            )

        parser = UnstructuredTextFileParser(self.mainfile, self._quantities)

        self._results = {
            key: parser[key] for key in parser.keys() if parser[key]}


class LammpsOutput:
    def __init__(self, filepath, archive, logger=None):
        self.filepath = filepath
        self.archive = archive
        self.logger = logger if logger is not None else logging
        self.log_parser = LogParser(filepath, self.logger)
        self.traj_parser = TrajParser(None, self.logger)
        self.data_parser = DataParser(None, self.logger)
        self._converter = None

    def parse_thermodynamic_data(self):
        thermo_data = self.log_parser.get_thermodynamic_data()
        if not thermo_data:
            return
        n_evaluations = len(thermo_data['Step'])

        energy_keys_mapping = {
            'e_pair': 'pair', 'e_bond': 'bond', 'e_angle': 'angle', 'e_dihed': 'dihedral',
            'e_impro': 'improper', 'e_coul': 'coulomb', 'e_vdwl': 'van der Waals',
            'e_mol': 'molecular', 'e_long': 'kspace long range',
            'e_tail': 'van der Waals long range', 'kineng': 'kinetic', 'poteng': 'potential',
        }

        sec_run = self.archive.section_run[-1]

        sec_sccs = sec_run.section_single_configuration_calculation

        create_scc = True
        if sec_sccs:
            if len(sec_sccs) != n_evaluations:
                self.logger.warn(
                    '''Mismatch in number of calculations %d and number of property
                    evaluations %d!, will create new sections''' % (
                        len(sec_sccs), n_evaluations))

            else:
                create_scc = False

        for n in range(n_evaluations):
            if create_scc:
                sec_scc = sec_run.m_create(section_single_configuration_calculation)
            else:
                sec_scc = sec_sccs[n]

            for key, val in thermo_data.items():
                key = key.lower()
                if key in energy_keys_mapping:
                    sec_energy = sec_scc.m_create(section_energy_contribution)
                    sec_energy.energy_contibution_kind = energy_keys_mapping[key]
                    sec_energy.energy_contribution_value = self._converter.Energy(val[n])

                elif key == 'toteng':
                    sec_scc.energy_method_current = self._converter.Energy(val[n])

                elif key == 'press':
                    sec_scc.pressure = self._converter.Press(val[n])

                elif key == 'temp':
                    sec_scc.temperature = self._converter.Temp(val[n])

                elif key == 'step':
                    sec_scc.time_step = int(val[n])

                elif key == 'cpu':
                    sec_scc.time_calculation = float(val[n])

                else:
                    if n == 0:
                        self.logger.warn(
                            'Unsupported property %s in thermodynamic data' % key)

    def parse_sampling_method(self):
        sec_run = self.archive.section_run[-1]
        sec_sampling_method = sec_run.m_create(section_sampling_method)

        run_style = self.log_parser.get('run_style', ['verlet'])[0]
        run = self.log_parser.get('run', [0])[0]
        timestep = self._converter.Time(self.log_parser.get('timestep', [0])[0])
        sampling_method, ensemble_type = self.log_parser.get_sampling_method()

        sec_sampling_method.x_lammps_integrator_type = run_style
        sec_sampling_method.x_lammps_number_of_steps_requested = run
        sec_sampling_method.x_lammps_integrator_dt = timestep
        sec_sampling_method.sampling_method = sampling_method
        sec_sampling_method.ensemble_type = ensemble_type

        thermo_settings = self.log_parser.get_thermostat_settings()
        target_T = thermo_settings.get('target_T', None)
        if target_T is not None:
            target_T = self._converter.Temp(target_T)
            sec_sampling_method.x_lammps_thermostat_target_temperature = target_T
        thermostat_tau = thermo_settings.get('thermostat_tau', None)
        if thermostat_tau is not None:
            thermostat_tau = self._converter.Time(thermostat_tau)
            sec_sampling_method.x_lammps_thermostat_tau = thermostat_tau
        target_P = thermo_settings.get('target_P', None)
        if target_P is not None:
            target_P = self._converter.Press(target_P)
            sec_sampling_method.x_lammps_barostat_target_pressure = target_P
        barostat_tau = thermo_settings.get('barostat_P', None)
        if barostat_tau is not None:
            barostat_tau = self._converter.Time(barostat_tau)
            sec_sampling_method.x_lammps_barostat_tau = barostat_tau
        langevin_gamma = thermo_settings.get('langevin_gamma', None)
        if langevin_gamma is not None:
            langevin_gamma = self._converter.Time(langevin_gamma)
            sec_sampling_method.x_lammps_langevin_gamma = langevin_gamma

    def parse_system(self, trajfile):
        sec_run = self.archive.section_run[-1]

        self.traj_parser.mainfile = trajfile

        pbc_cell = self.traj_parser.get('pbc_cell', [])
        n_atoms = self.traj_parser.get('n_atoms', [])

        create_scc = True
        sec_sccs = sec_run.section_single_configuration_calculation
        if sec_sccs:
            if len(sec_sccs) != len(pbc_cell):
                self.logger.warn(
                    '''Mismatch in number of calculations %d and number of property
                    evaluations %d!, will create new sections''' % (
                        len(sec_sccs), len(pbc_cell)))

            else:
                create_scc = False

        for i in range(len(pbc_cell)):
            sec_system = sec_run.m_create(section_system)
            sec_system.number_of_atoms = n_atoms[i]
            sec_system.configuration_periodic_dimensions = pbc_cell[i][0]
            sec_system.simulation_cell = self._converter.Distance(pbc_cell[i][1])
            sec_system.atom_positions = self.traj_parser.get_positions(i)
            atom_labels = self.traj_parser.get_atom_labels(i)
            if atom_labels is None:
                atom_labels = ['X'] * n_atoms[i]
            sec_system.atom_labels = atom_labels

            velocities = self.traj_parser.get_velocities(i)
            if velocities is not None:
                sec_system.atom_velocities = self._converter.Velocity(velocities)

            forces = self.traj_parser.get_forces(i)
            if forces is not None:
                if create_scc:
                    sec_scc = sec_run.m_create(section_single_configuration_calculation)
                else:
                    sec_scc = sec_sccs[i]

                sec_scc.atom_forces = self._converter.Force(forces)

    def parse_topology(self, datafile):
        sec_run = self.archive.section_run[-1]

        self.data_parser.mainfile = datafile

        masses = self.data_parser.get('Masses', None)

        self.traj_parser.masses = masses

        sec_topology = sec_run.m_create(section_topology)
        sec_topology.number_of_topology_atoms = self.data_parser.get('atoms', [None])[0]

        interactions = self.log_parser.get_interactions()
        if not interactions:
            interactions = self.data_parser.get_interactions()

        for interaction in interactions:
            if not interaction[0] or not interaction[1]:
                continue
            sec_interaction = sec_topology.m_create(section_interaction)
            sec_interaction.interaction_kind = str(interaction[0])
            sec_interaction.interaction_parameters = [list(a) for a in interaction[1]]

    def parse(self):
        sec_run = self.archive.m_create(section_run)
        # TODO Warning! lj units are not supported in LammpsCommon!
        self._converter = converter(self.log_parser.get('units', ['lj'])[0])

        # parse basic
        sec_run.program_name = 'LAMMPS'
        sec_run.program_version = self.log_parser.get_program_version()

        # parse method-related
        self.parse_sampling_method()

        # parse all data files associated with calculation, normally only one
        # specified in log file, otherwise will scan directory
        data_files = self.log_parser.get_data_files()
        for data_file in data_files:
            self.parse_topology(data_file)

        # parse all trajectorty files associated with calculation, normally only one
        # specified in log file, otherwise will scan directory
        traj_files = self.log_parser.get_traj_files()
        for traj_file in traj_files:
            self.parse_system(traj_file)

        # parse thermodynamic data from log file
        self.parse_thermodynamic_data()

        # create workflow
        if sec_run.section_sampling_method[0].sampling_method:
            sec_workflow = self.archive.m_create(Workflow)
            sec_workflow.workflow_type = sec_run.section_sampling_method[0].sampling_method

            if sec_workflow.workflow_type == 'molecular_dynamics':
                sec_md = sec_workflow.m_create(MolecularDynamics)

                sec_md.finished_normally = self.log_parser.finished_normally()
                sec_md.with_trajectory = self.traj_parser.with_trajectory()
                sec_md.with_thermodynamics = self.log_parser.with_thermodynamics()



# Copyright 2016-2018 Sami Kivistö, Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.simple_parser import extractOnCloseTriggers

META_PREFIX = "x_crystal_input"
META_SECTION_PREFIX = "x_crystal_section_input"


class InputMatcher(SM):
    """A subclass of SimpleMatcher that makes the handling of input a bit
    simpler.
    """
    def __init__(self,
            startReStr,
            endReStr=None,
            subMatchers=tuple(),
            sections=tuple(),
            subFlags=SM.SubFlags.Sequenced,
            weak=False,     # this matcher should not "steal" the position
            repeats=False,  # this matcher is expected to repeat
            required=False,  # this value is required to have been matched on section close
            floating=False,  # this section goes not steal the context, useful for example for low level debugging/error messages.
            # valid from the point they are read until the exit from the enclosing section
            forwardMatch=False,  # if start match should not eat input, but be forwarded to adHoc and subMatchers
            name="",
            adHoc=None,
            otherMetaInfo=tuple(),  # The metainfos that are later manually added ot the backend
            fixedStartValues=None,
            fixedEndValues=None,
            dependencies=None,
            defLine=0,
            defFile='',
            coverageIgnore=False,  # mark line as ignored in coverage analysis
            onClose=None,   # A dictionary of onClose callbacks that are specific to this SimpleMatcher
            onOpen=None,   # A dictionary of onOpen callbacks that are specific to this SimpleMatcher
            startReAction=None,  # A callback function that is called with the groups that were matched from the startReStr.
            ):

        if name != "":
            startReStr = r"(?i)(?P<{}_{}>{})".format(META_PREFIX, name, startReStr)

        for i_section, section in enumerate(sections):
            if section:
                sections[i_section] = "{}_{}".format(META_SECTION_PREFIX, section)
            else:
                sections[i_section] = META_SECTION_PREFIX

        super(InputMatcher, self).__init__(
            startReStr,
            endReStr,
            subMatchers,
            sections,
            subFlags,
            weak,
            repeats,
            required,
            floating,
            forwardMatch,
            name,
            adHoc,
            otherMetaInfo,
            fixedStartValues,
            fixedEndValues,
            dependencies,
            defLine,
            defFile,
            coverageIgnore,
            onClose,
            onOpen,
            startReAction,
        )


class CrystalInputParser(object):

    def __init__(self, mainparser):

        # Gather onClose triggers that are defined within this class and add
        # them to the main parser
        on_close = {}
        for attr, callback in extractOnCloseTriggers(self).items():
            on_close[attr] = [callback]
        mainparser.on_close.update(on_close)

        IM = InputMatcher

        # Define root matcher
        self.root_matcher = IM("",
            forwardMatch=True,
            sections=[""],
            subMatchers=[
                IM("CRYSTAL|SLAB|POLYMER|HELIX|MOLECULE|EXTERNAL|DLVINPUT", name="system_type"),
                IM("OPTGEOM|FREQCALC|ANHARM", name="calculation_type"),
                IM("DFT",
                    sections=["dft"],
                    subFlags=SM.SubFlags.Unordered,
                    subMatchers=[
                        IM("SPIN"),
                        IM("EXCHANGE",
                            subMatchers=[
                                IM("LDA|VBH|BECKE|PBE|PBESOL|mPW91|PWGGA|SOGGA|WCGGA", name="dft_exchange"),
                            ]
                        ),
                        IM("CORRELAT",
                            subMatchers=[
                                IM("PZ|VBH|VWN|LYP|P86|PBE|PBESOL|PWGGA|PWLSD|WL", name="dft_correlation")
                            ]
                        ),
                        IM("SVWN|BLYP|PBEXC|PBESOLXC|SOGGAXC|B3PW|B3LYP|PBE0|PBESOL0|B1WC|WCILYP|B97H|PBE0-13|HYBRID|NONLOCAL|HSE06|HSESOL|HISS|RSHXLDA|wB97|wB97X|LC-WPBE|LC-WPBESOL|LC-WBLYP|M05-2X|M05|M062X|M06HF|M06L|M06|B2PLYP|B2GPPLYP|mPW2PLYP|DHYBRID", name="dft_xc_shortcut")
                    ]
                ),
            ]
        )

    def onClose_x_crystal_section_input_dft(self, backend, gIndex, section):
        shortcut = section.get_latest_value("x_crystal_input_dft_xc_shortcut")
        exchange = section.get_latest_value("x_crystal_input_dft_exchange")
        correlation = section.get_latest_value("x_crystal_input_dft_correlation")

        xc_list = []
        xc_summary = []

        # Handle the XC's defined with single shortcut
        if shortcut:
            shortcut = shortcut.upper()
            shortcut_map = {
                "PBE0": "HYB_GGA_XC_PBEH",
                "B3LYP": "HYB_GGA_XC_B3LYP",
                "HSE06": "HYB_GGA_XC_HSE06",
                "M06": "HYB_MGGA_XC_M06",
                "M05-2X": "HYB_MGGA_XC_M05_2X",
                "LC-WPBE": "HYB_GGA_XC_LRC_WPBE",
            }
            norm_xc = shortcut_map.get(shortcut)
            if norm_xc:
                xc_list.append(norm_xc)

        # Handle the exchange part
        if exchange:
            exchange = exchange.upper()
            exchange_map = {
                "PBE": "GGA_X_PBE",
                "PBESOL": "GGA_X_PBE_SOL",
                "BECKE": "GGA_X_B88",
                "LDA": "LDA_X",
                "PWGGA": "GGA_X_PW91",
            }
            norm_x = exchange_map.get(exchange)
            if norm_x:
                xc_list.append(norm_x)

        # Handle the correlation part
        if correlation:
            correlation = correlation.upper()
            correlation_map = {
                "PBE": "GGA_C_PBE",
                "PBESOL": "GGA_C_PBE_SOL",
                "PZ": "LDA_C_PZ",
                "WFN": "LDA_C_VWN",
                "PWGGA": "GGA_C_PW91",
            }
            norm_c = correlation_map.get(correlation)
            if norm_c:
                xc_list.append(norm_c)

        # Go throught the XC list and add the sections and gather a summary
        for xc in xc_list:
            sid = backend.openSection("section_XC_functionals")
            weight = 1.0
            backend.addValue("XC_functional_name", xc)
            backend.addValue("XC_functional_weight", weight)
            backend.closeSection("section_XC_functionals", sid)
            xc_summary.append("{}*{}".format(weight, xc))

        if len(xc_list) == 0:
            print(shortcut)
            print(exchange)
            print(correlation)

        backend.addValue("XC_functional", "+".join(sorted(xc_summary)))

    def debug(self):
        print("DEBUG")



import os
from glob import glob
import unittest

from lark import Tree

from optimade.filterparser import LarkParser, ParserError

testfile_dir = os.path.join(os.path.dirname(__file__), "testfiles")


class ParserTestV0_9_5(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_filters = []
        for fn in sorted(glob(os.path.join(testfile_dir, "*.inp"))):
            with open(fn) as f:
                cls.test_filters.append(f.read().strip())

    def setUp(self):
        self.parser = LarkParser(version=(0, 9, 5))

    def test_inputs(self):
        for tf in self.test_filters:
            if tf == "filter=number=0.0.1":
                self.assertRaises(ParserError, self.parser.parse, tf)
            else:
                tree = self.parser.parse(tf)
                self.assertTrue(tree, Tree)

    def test_parser_version(self):
        v = (0, 9, 5)
        p = LarkParser(version=v)
        self.assertIsInstance(p.parse(self.test_filters[0]), Tree)
        self.assertEqual(p.version, v)

    def test_repr(self):
        self.assertIsNotNone(repr(self.parser))
        self.parser.parse(self.test_filters[0])
        self.assertIsNotNone(repr(self.parser))


class ParserTestV0_10_1(unittest.TestCase):
    version = (0, 10, 1)
    variant = "default"

    @classmethod
    def setUpClass(cls):
        cls.parser = LarkParser(version=cls.version, variant=cls.variant)

    def parse(self, inp):
        return self.parser.parse(inp)

    def test_empty(self):
        self.assertIsInstance(self.parse(" "), Tree)

    def test_property_names(self):
        self.assertIsInstance(self.parse("band_gap = 1"), Tree)
        self.assertIsInstance(self.parse("cell_length_a = 1"), Tree)
        self.assertIsInstance(self.parse("cell_volume = 1"), Tree)

        with self.assertRaises(ParserError):
            self.parse("0_kvak IS KNOWN")  # starts with a number

        with self.assertRaises(ParserError):
            self.parse('"foo bar" IS KNOWN')  # contains space; contains quotes

        with self.assertRaises(ParserError):
            self.parse("BadLuck IS KNOWN")  # contains upper-case letters

        # database-provider-specific prefixes
        self.assertIsInstance(self.parse("_exmpl_formula_sum = 1"), Tree)
        self.assertIsInstance(self.parse("_exmpl_band_gap = 1"), Tree)

        # Nested property names
        self.assertIsInstance(self.parse("identifier1.identifierd2 = 42"), Tree)

    def test_string_values(self):
        self.assertIsInstance(self.parse('author="Sąžininga Žąsis"'), Tree)
        self.assertIsInstance(
            self.parse('field = "!#$%&\'() * +, -./:; <= > ? @[] ^ `{|}~ % "'), Tree
        )

    def test_number_values(self):
        self.assertIsInstance(self.parse("a = 12345"), Tree)
        self.assertIsInstance(self.parse("b = +12"), Tree)
        self.assertIsInstance(self.parse("c = -34"), Tree)
        self.assertIsInstance(self.parse("d = 1.2"), Tree)
        self.assertIsInstance(self.parse("e = .2E7"), Tree)
        self.assertIsInstance(self.parse("f = -.2E+7"), Tree)
        self.assertIsInstance(self.parse("g = +10.01E-10"), Tree)
        self.assertIsInstance(self.parse("h = 6.03e23"), Tree)
        self.assertIsInstance(self.parse("i = .1E1"), Tree)
        self.assertIsInstance(self.parse("j = -.1e1"), Tree)
        self.assertIsInstance(self.parse("k = 1.e-12"), Tree)
        self.assertIsInstance(self.parse("l = -.1e-12"), Tree)
        self.assertIsInstance(self.parse("m = 1000000000.E1000000000"), Tree)

        with self.assertRaises(ParserError):
            self.parse("number=1.234D12")
        with self.assertRaises(ParserError):
            self.parse("number=.e1")
        with self.assertRaises(ParserError):
            self.parse("number= -.E1")
        with self.assertRaises(ParserError):
            self.parse("number=+.E2")
        with self.assertRaises(ParserError):
            self.parse("number=1.23E+++")
        with self.assertRaises(ParserError):
            self.parse("number=+-123")
        with self.assertRaises(ParserError):
            self.parse("number=0.0.1")

    def test_operators(self):
        # Basic boolean operations
        self.assertIsInstance(
            self.parse(
                'NOT ( chemical_formula_hill = "Al" AND chemical_formula_anonymous = "A" OR '
                'chemical_formula_anonymous = "H2O" AND NOT chemical_formula_hill = "Ti" )'
            ),
            Tree,
        )

        # Numeric and String comparisons
        self.assertIsInstance(self.parse("nelements > 3"), Tree)
        self.assertIsInstance(
            self.parse(
                'chemical_formula_hill = "H2O" AND chemical_formula_anonymous != "AB"'
            ),
            Tree,
        )
        self.assertIsInstance(
            self.parse(
                "_exmpl_aax <= +.1e8 OR nelements >= 10 AND "
                'NOT ( _exmpl_x != "Some string" OR NOT _exmpl_a = 7)'
            ),
            Tree,
        )
        self.assertIsInstance(self.parse('_exmpl_spacegroup="P2"'), Tree)
        self.assertIsInstance(self.parse("_exmpl_cell_volume<100.0"), Tree)
        self.assertIsInstance(
            self.parse("_exmpl_bandgap > 5.0 AND _exmpl_molecular_weight < 350"), Tree
        )
        self.assertIsInstance(
            self.parse('_exmpl_melting_point<300 AND nelements=4 AND elements="Si,O2"'),
            Tree,
        )
        self.assertIsInstance(self.parse("_exmpl_some_string_property = 42"), Tree)
        self.assertIsInstance(self.parse("5 < _exmpl_a"), Tree)

        # OPTIONAL
        self.assertIsInstance(
            self.parse("((NOT (_exmpl_a>_exmpl_b)) AND _exmpl_x>0)"), Tree
        )
        self.assertIsInstance(self.parse("5 < 7"), Tree)

    def test_id(self):
        self.assertIsInstance(self.parse('id="example/1"'), Tree)
        self.assertIsInstance(self.parse('"example/1" = id'), Tree)
        self.assertIsInstance(self.parse('id="test/2" OR "example/1" = id'), Tree)

    def test_string_operations(self):
        #  Substring comparisons
        self.assertIsInstance(
            self.parse(
                'chemical_formula_anonymous CONTAINS "C2" AND '
                'chemical_formula_anonymous STARTS WITH "A2"'
            ),
            Tree,
        )
        self.assertIsInstance(
            self.parse(
                'chemical_formula_anonymous STARTS "B2" AND '
                'chemical_formula_anonymous ENDS WITH "D2"'
            ),
            Tree,
        )

    def test_list_properties(self):
        # Comparisons of list properties
        self.assertIsInstance(self.parse("list HAS < 3"), Tree)
        self.assertIsInstance(self.parse("list HAS ALL < 3, > 3"), Tree)
        self.assertIsInstance(self.parse("list:list HAS >=2:<=5"), Tree)
        self.assertIsInstance(
            self.parse(
                'elements HAS "H" AND elements HAS ALL "H","He","Ga","Ta" AND elements HAS '
                'ONLY "H","He","Ga","Ta" AND elements HAS ANY "H", "He", "Ga", "Ta"'
            ),
            Tree,
        )

        # OPTIONAL:
        self.assertIsInstance(self.parse('elements HAS ONLY "H","He","Ga","Ta"'), Tree)
        self.assertIsInstance(self.parse('elements HAS ALL "H","He","Ga","Ta"'), Tree)
        self.assertIsInstance(self.parse('elements HAS ANY "H","He","Ga","Ta"'), Tree)
        self.assertIsInstance(
            self.parse(
                'elements:_exmpl_element_counts HAS "H":6 AND '
                'elements:_exmpl_element_counts HAS ALL "H":6,"He":7 AND '
                'elements:_exmpl_element_counts HAS ONLY "H":6 AND '
                'elements:_exmpl_element_counts HAS ANY "H":6,"He":7 AND '
                'elements:_exmpl_element_counts HAS ONLY "H":6,"He":7'
            ),
            Tree,
        )
        self.assertIsInstance(
            self.parse(
                "_exmpl_element_counts HAS < 3 AND "
                "_exmpl_element_counts HAS ANY > 3, = 6, 4, != 8"
            ),
            Tree,
        )
        self.assertIsInstance(
            self.parse(
                "elements:_exmpl_element_counts:_exmpl_element_weights "
                'HAS ANY > 3:"He":>55.3 , = 6:>"Ti":<37.6 , 8:<"Ga":0'
            ),
            Tree,
        )

    def test_properties(self):
        #  Filtering on Properties with unknown value
        self.assertIsInstance(
            self.parse(
                "chemical_formula_hill IS KNOWN AND "
                "NOT chemical_formula_anonymous IS UNKNOWN"
            ),
            Tree,
        )

    def test_precedence(self):
        self.assertIsInstance(self.parse('NOT a > b OR c = 100 AND f = "C2 H6"'), Tree)
        self.assertIsInstance(
            self.parse('(NOT (a > b)) OR ( (c = 100) AND (f = "C2 H6") )'), Tree
        )
        self.assertIsInstance(self.parse("a >= 0 AND NOT b < c OR c = 0"), Tree)
        self.assertIsInstance(
            self.parse("((a >= 0) AND (NOT (b < c))) OR (c = 0)"), Tree
        )

    def test_special_cases(self):
        self.assertIsInstance(self.parse("te < st"), Tree)
        self.assertIsInstance(self.parse('spacegroup="P2"'), Tree)
        self.assertIsInstance(self.parse("_cod_cell_volume<100.0"), Tree)
        self.assertIsInstance(
            self.parse("_mp_bandgap > 5.0 AND _cod_molecular_weight < 350"), Tree
        )
        self.assertIsInstance(
            self.parse('_cod_melting_point<300 AND nelements=4 AND elements="Si,O2"'),
            Tree,
        )
        self.assertIsInstance(self.parse("key=value"), Tree)
        self.assertIsInstance(self.parse('author=" someone "'), Tree)
        self.assertIsInstance(self.parse('author=" som\neone "'), Tree)
        self.assertIsInstance(
            self.parse(
                "number=0.ANDnumber=.0ANDnumber=0.0ANDnumber=+0AND_n_u_m_b_e_r_=-0AND"
                "number=0e1ANDnumber=0e-1ANDnumber=0e+1"
            ),
            Tree,
        )

        self.assertIsInstance(
            self.parse("NOTice=val"), Tree
        )  # property (ice) != property (val)
        self.assertIsInstance(
            self.parse('NOTice="val"'), Tree
        )  # property (ice) != value ("val")
        self.assertIsInstance(
            self.parse('"NOTice"=val'), Tree
        )  # value ("NOTice") = property (val)

        with self.assertRaises(ParserError):
            self.parse("NOTICE=val")  # not valid property or value (NOTICE)
        with self.assertRaises(ParserError):
            self.parse('"NOTICE"=Val')  # not valid property (Val)
        with self.assertRaises(ParserError):
            self.parse("NOTICE=val")  # not valid property or value (NOTICE)

    def test_parser_version(self):
        self.assertEqual(self.parser.version, self.version)
        self.assertEqual(self.parser.variant, self.variant)

    def test_repr(self):
        self.assertIsNotNone(repr(self.parser))
        self.parser.parse('key="value"')
        self.assertIsNotNone(repr(self.parser))
